/*
 * Some simple Read/Write functions
 */

#include "pexec.h"
#include <errno.h>

/*@
   PX_Read - Basic read function

   Input Parameters:
.  fd  - File descriptor to read
.  buf - Data area to read into
.  len - Maximum amount of data to read
.  ctx - Pointer to PX_ReadFn that holds this function (self)

   Returns:
   Number of bytes read, or -1 if there is an error.  If the read
   was aborted because of an interrupted system call (EINTR), the
   read is restarted. 
@*/
int PX_Read( fd, buf, len, ctx )
int       fd, len;
void      *buf;
PX_ReadFn ctx;
{
    int n;
    do {
	n = read( fd, buf, len );
    } while (n < 0 && errno == EINTR);
    return n;
}

/*@
   PX_Write - Basic write function

   Input Parameters:
.  fd  - File descriptor to write
.  buf - Data area to write
.  len - Amount of data to write
.  ctx - Pointer to PX_WriteFn that holds this function (self)

   Returns:
   Number of bytes written, or -1 if there is an error.  If the write
   was aborted because of an interrupted system call (EINTR), the
   write is restarted. 
@*/
int PX_Write( fd, buf, len, ctx )
int        fd, len;
void       *buf;
PX_WriteFn ctx;
{
    int n;
    do {
	n = write( fd, buf, len );
	if (n < 0 && errno == EINTR) continue;
	} while (0);
    return n;
}

/* 
 * Need routines to insert/delete read/write handlers.
 */

/*@
  PX_Read_add - Adds a function for reading from an FD

  Input Parameters:
. ctx - PEXEC context
. fd  - File descriptor to read from
. read - Function to call for actual read
. extra_data - Extra data saved for function

  Notes:
  By default, the function 'PX_Read' is called to perform reads.  This function
  allows the user to choose a different routine or to interpose others.
  The read function is added at the `head` of the list; it may call functions
  further down in the list to complete the operation.  See the tracing 
  examples for more information.

@*/
int PX_Read_add( ctx, fd, read, extra_data )
PX_Ctx *ctx;
int    fd;
int    read ANSI_ARGS((int,void *,int,PX_ReadFn));
void   *extra_data;
{
    PX_ReadFn rfn;

    rfn = (PX_ReadFn) MALLOC( sizeof(struct _PX_ReadFn) );
    if (!rfn) return -1;
    rfn->fcn	   = read;
    rfn->extra_data	   = extra_data;
    
    rfn->next	   = ctx->conn[fd].read;
    rfn->prev	   = 0;
    ctx->conn[fd].read = rfn;
    
    return 0;
}

/*@
  PX_Write_add - Adds a function for writing to an FD

  Input Parameters:
. ctx - PEXEC context
. fd  - File descriptor to write to
. read - Function to call for actual write
. extra_data - Extra data saved for function

  Notes:
  By default, the function 'PX_Write' is called to perform writes.
  This function
  allows the user to choose a different routine or to interpose others.
  The write function is added at the `head` of the list; it may call functions
  further down in the list to complete the operation.  See the tracing 
  examples for more information.

@*/
int PX_Write_add( ctx, fd, write, extra_data )
PX_Ctx *ctx;
int    fd;
int    write ANSI_ARGS((int,void *,int,PX_WriteFn));
void   *extra_data;
{
    PX_WriteFn wfn;
    
    wfn = (PX_WriteFn) MALLOC( sizeof(struct _PX_WriteFn) );
    if (!wfn) return -1;
    wfn->fcn	    = write;
    wfn->extra_data	    = extra_data;

    wfn->next	    = ctx->conn[fd].write;
    wfn->prev	    = 0;
    ctx->conn[fd].write = wfn;
    
    return 0;
}

/* 
 * This should be a general structure for both nonblocking reads and writes; 
 * we should consider some preallocation of it.
 */
typedef struct {
    void *buf;
    char *cbuf;
    int  nleft;
    void (*complete_fcn) ANSI_ARGS((void *, void *, int));
    void *fcn_ctx;
    } WriteNbData;

/* Internal write handler */
int PX_Hfcn_Write_nb( ctx, conn, handler, n )
PX_Ctx     *ctx;
PX_Conn    *conn;
PX_Handler handler;
int        n;
{
    WriteNbData *ndata = (WriteNbData *)handler->extra_data;
    int n_written;

    /* Not correct; this is US */
    /* if (conn->write_pending) return 0; */

    n_written = PX_SYSWRITE( conn, ndata->cbuf, ndata->nleft );
    if (n_written > 0) {
	ndata->nleft -= n_written;
	if (ndata->nleft == 0) {
	    /* Done! */
	    (*ndata->complete_fcn)( ndata->buf, ndata->fcn_ctx, 0 );
	    /* Remove this handler */
	    return 0;
	    }
	/* Not done reading */
	return 0;
	}
    else {
        (*ndata->complete_fcn)( ndata->buf, ndata->fcn_ctx, n_written );
	return 1;
	}
}
/* We could preallocate a list of these... */
/*
   PX_Write_nb - Nonblocking write routine for PEXEC

   Notes:
   This routine tries to write the data directly.  If not all of the data can
   be sent, a write handler is added to the connection to send the data.

   ? Must just add handler if one is already present

   Returns:
   Number of bytes written.  -1 on error.

   The completion function is called when all data has been written.
 */
int PX_Write_nb( conn, buf, n, complete_fcn, fcn_ctx, deltasecs )
PX_Conn *conn;
void    *buf, *fcn_ctx;
int     n, deltasecs;
void    (*complete_fcn) ANSI_ARGS((void *, void *, int));
{
    int         n_written = 0;
    WriteNbData *ndata;

    if (!conn->write_pending)
	n_written = PX_SYSWRITE( conn, buf, n );
    if (n_written >= 0) {
	if (n_written == n) {
	    (*complete_fcn)( buf, fcn_ctx, 0 ); /* n_written? */
	    return n_written;
	    }
	/* Only some data sent. Save the information on the rest, install
	   a handler, and exit */
	ndata = (WriteNbData *)MALLOC( sizeof(WriteNbData) );
	if (!ndata) return -1;
	ndata->cbuf         = ((char *)buf) + n_written;
	ndata->nleft        = n - n_written;
	ndata->complete_fcn = complete_fcn;
	ndata->fcn_ctx      = fcn_ctx;
/*	write_handler = PX_Handler_simple_create( ); */
	}
    else {
	/* Error case */
	; 
	}
    return n_written;
}

/* 
   We should have the same kind of read routine, which reads a given number of
   bytes and then calls a specified routine.
 */
