/*
 * This file contains the necessary code to start a single, remote job
 * using a remote shell command (usually rsh).  
 * Handlers are provided by the user; an example "handshake" version
 * is defined here which allows cooperating tasks to confirm that
 * the requested process has started.  This is necessary because some
 * implementations of rsh/rshd will hang insted of failing; providing the
 * handshake allows us to detect this failure mode.
 */

#include "pexec.h"
#include <memory.h>
#include <stdio.h>

#define MAX_HANDSHAKE 128
typedef struct {
    char handshake[MAX_HANDSHAKE+1];
    int  len;       /* Total length of handshake */
    int  loc;       /* Number of matched characters */
    } PX_Handshake;

/*@
   PX_Create_handshake - Create a handshake structure

   Input Parameter:
.  str - String to use as the handshake

   Returns:
   Pointer to 'extra_data' to give when creating handler with 
   'PX_Hfcn_handshake'.  A typical use is
.vb
   PX_Handler_simple_create( PX_Hfcn_handshake, 
                      PX_Create_handshake( "My_String" ) );
.ve

   Notes:
   The routine at the other end should write the string to the other side of
   the pipe/socket that this handler is connected to.

 @*/
void *PX_Create_handshake( str )
char *str;
{
PX_Handshake *new;

new = (PX_Handshake *)MALLOC( sizeof( PX_Handshake ) );
if (!new) return 0;
new->len = strlen( str );
if (new->len > MAX_HANDSHAKE) {
    fprintf( stderr, "Handshake string may not be longer than %d\n",
	     MAX_HANDSHAKE );
    FREE( new );
    return 0;
    }
new->loc = 0;
strcpy( new->handshake, str );
return (void *) new;
}

/*@
  PX_Hfcn_handshake - Handler for handshake code

  Notes:
  This implements a simple handshake on a connection.  This should be
  combined with a reasonable timeout to handle failure of the handshake.

.seealso: PX_Create_handshake
@*/
int PX_Hfcn_handshake( ctx, conn, handler, in_out )
PX_Ctx     *ctx;
PX_Conn    *conn;
PX_Handler handler;
int        in_out;
{
PX_Handshake *handshake = (PX_Handshake *)handler->extra_data;
int err = 0;
int i, n;
char buf[MAX_HANDSHAKE], *p1;

n = PX_SYSREAD( conn, buf, handshake->len-handshake->loc );
if (n <= 0) 
    return PX_Hfcn_do_eof( ctx, conn, handler, n );

/* Compare what we've read with the handshake string.  Either advance
   "loc" or go to the error state */
p1 = handshake->handshake + handshake->loc;
for (i=0; i<n; i++) {
    if (buf[i] != *p1++) {
	/* Error - handshake does not match */
	(*conn->on_err->fcn)( ctx, conn, conn->on_err, i + handshake->loc );
	return -1;
	}
    }
handshake->loc += n;
if (handshake->loc == handshake->len) {
    /* Got handshake; install new handlers - from the "next" field */
    PX_FD_Remove_handler( conn, handler );
    }

return 0;
}

/*@
  PX_Proc_run_on_host - Runs a program on a host, using remote shell

  Input Parameters:
. host - Name of host
. pgm  - Name of program
. argv - Argument vector; null terminated

  Output Parameters:
. rd_in - fd for standard input to program
. wrt_out - fd for standard output from program
. wrt_err - fd for standard error from program
. pgmst  - Process description structure

 Notes:
 This process returns as soon as the fork to create the process to run the
 remote shell in completes.  Robust applications should consider a
 'PX_Hfcn_handshake' handler (on wrt_out) and a timeout.
@*/
int PX_Proc_run_on_host( host, pgm, argv, rd_in, wrt_out, wrt_err, pgmst )
char *host, *pgm, **argv;
int  *rd_in, *wrt_out, *wrt_err;
PX_Pgm **pgmst;
{
char *rpgm, **rargv;
int  narg, i;
/* Synthesis a NEW argv/pgm for the remote shell command */

narg = 0;
while (argv[narg]) narg++;
narg++;

rargv = (char **)MALLOC( (narg + 4) * sizeof(char **) );
if (!rargv) {
    return -1;
    }
/* Question: should we COPY the arguments, or just the pointers?  
   Copy is probably better, just to keep track of things */
for (i=0; i<narg; i++) {
    rargv[i+2] = argv[i];
    }
rargv[0] = (char *)MALLOC( strlen(RSHCOMMAND)+1 );
if (!rargv[0]) return -1;
strcpy( rargv[0], RSHCOMMAND );
rpgm = rargv[0];

rargv[1] = (char *)MALLOC( strlen(host) + 1 );
if (!rargv[1]) return -1;
strcpy( rargv[1], host );

return PX_Proc_run_program( rpgm, rargv, rd_in, wrt_out, wrt_err, pgmst );
}
