#include "pexec.h"
#include <errno.h>
#include <sys/param.h>
#include <sys/socket.h>

/* sockaddr_in (Internet) */
#include <netinet/in.h>
/* TCP_NODELAY */
#include <netinet/tcp.h>

/* sockaddr_un (Unix) */
#include <sys/un.h>

/* defs of gethostbyname */
#include <netdb.h>

/* fcntl, F_GET/SETFL */
#include <fcntl.h>

#ifndef MAX_HOST_NAME
#define MAX_HOST_NAME 1024
#endif

/* This is really IP!? */
#ifndef TCP
#define TCP 0
#endif

/* 
 This structure is used to keep track of the connection information. 
 Accept and Connect calls may need this.
 */
typedef struct {
    int                sock_type;    /* AF_UNIX or AF_INET */
    struct sockaddr_in s1;
    struct sockaddr_un s2;
    int                slen;         /* Length of s1 or s2, depending on 
					sock_type */
    int                authenticated; /* whether or not user is OK */
    } PX_Socket_info;

/*@
  PX_FD_Create_network_socket - Creates a file descriptor for a network (INET)
  socket

  Input Parameters:
. hostname - Name of host; network address is (portnum,hostname).  May be
  null, in which case 'gethostname' is used.
. portnum - Port number.
. max_number_of_pending_conn - Controls the number of pending connections.
  A reasonable value is zero.

  Output Parameter:
. sock_info - Pointer to data to give 'PX_FD_Accept_connection'.  Use
  'PX_FD_Free_sockinfo' when closing this FD.

  Returns:
  A file descriptor or '-1'.  

  Notes:
  This creates an fd that a 'listen' has been performed on.  The routine, 
  'PX_FD_Accept_connection' should be called to create an fd for communication
  with another process.  That process should use 'PX_FD_Create_network_conn'
  and 'PX_FD_Establish_network_conn' to establish the other side of the 
  connection. 

  In Master/Slave or Server/Client applications, this routine is used by the
  Master/Server.
@*/
int PX_FD_Create_network_socket( hostname, portnum, 
				 max_number_of_pending_conn, sock_info )
int  portnum, max_number_of_pending_conn;
char *hostname;
void **sock_info;
{
int                fd;
struct sockaddr_in sa;
struct hostent     *hp;
char               myname[MAX_HOST_NAME+1];
int                optval = 1;
PX_Socket_info     *sock;

/*
 * First, get network name if necessary
 */
if (!hostname) {
    hostname = myname;
    gethostname( myname, MAX_HOST_NAME );
    }
hp = gethostbyname( hostname );
if (!hp) {
    return -1;
    }

sock = (PX_Socket_info *)MALLOC( sizeof(PX_Socket_info) );
if (!sock) return -1;

bzero( &sa, sizeof(sa) );
sa.sin_family	   = AF_INET;
sa.sin_port	   = htons( portnum );
sa.sin_addr.s_addr = INADDR_ANY;

fd = socket( AF_INET, SOCK_STREAM, TCP );
if (fd < 0) {
    return fd;
    }

if (setsockopt( fd, IPPROTO_TCP, TCP_NODELAY, 
	        (char *)&optval, sizeof(optval) )) {
    perror( "Error calling setsockopt:" );
    }

if (bind( fd, (struct sockaddr *)&sa, sizeof(sa) ) < 0) {
    close( fd );
    return -1;
/*
    possible errors are:
EADDRINUSE
EADDRNOTAVAIL
 */
    }

/* Initialize the socket info fields */
sock->sock_type	    = AF_INET;
sock->s1	    = sa;
sock->slen	    = sizeof(sa);
sock->authenticated = 0;

/* Listen is a non-blocking call that enables connections */
listen( fd, max_number_of_pending_conn );

*sock_info = (void *)sock;

return fd;
}

/*@
  PX_FD_Create_local_socket - Creates a file descriptor for a local (Unix 
  domain) socket

  Input Parameters:
. name - Valid name in the file system for the local socket.
. max_number_of_pending_conn - Controls the number of pending connections.
  A reasonable value is zero.

  Output Parameter:
. sock_info - Pointer to data to give 'PX_FD_Accept_connection'.  Use
  'PX_FD_Free_sockinfo' when closing this FD.

  Returns:
  A file descriptor or '-1'.  

  Notes:
  This creates an fd that a 'listen' has been performed on.  The routine, 
  'PX_FD_Accept_connection' should be called to create an fd for communication
  with another process.  That process should use 'PX_FD_Create_local_conn'
  and 'PX_FD_Establish_local_conn' to establish the other side of the 
  connection. 

  In Master/Slave or Server/Client applications, this routine is used by the
  Master/Server.
@*/
int PX_FD_Create_local_socket( name, max_number_of_pending_conn, sock_info )
char *name;
int  max_number_of_pending_conn;
void **sock_info;
{
int                fd;
struct sockaddr_un sa;
PX_Socket_info     *sock;

sock = (PX_Socket_info *)MALLOC( sizeof(PX_Socket_info) );
if (!sock) return -1;

bzero( &sa, sizeof(sa) );
sa.sun_family = AF_UNIX;
strncpy( sa.sun_path, name, sizeof(sa.sun_path)-1 );

fd = socket( AF_UNIX, SOCK_STREAM, TCP );
if (fd < 0) {
    return fd;
    }

if (bind( fd, (struct sockaddr *)&sa, sizeof(sa) ) < 0) {
    close( fd );
    return -1;
/*
    possible errors are:
EADDRINUSE
EADDRNOTAVAIL
EIO/EISDIR/ELOOP/ENAMETOOLONG/ENOENT/ENOTDIR/EROFS for file name 
name must be manually unlinked on close
 */
    }
/* Initialize the socket info fields */
sock->sock_type = AF_UNIX;
sock->s2        = sa;
sock->slen      = sizeof(sa);

/* Listen is a non-blocking call that enables connections */
listen( fd, max_number_of_pending_conn );

*sock_info = (void *)sock;

return fd;
}

/* 
   This routine should be called when there is activity on a listen fd.
   It creates a NEW fd that gives a socket to the partner.
   That fd needs to be installed into the select context (with the appropriate
   handler) 
   
   This is NOT a handler; rather, it is a routine that handlers call.
 */
/*@
  PX_FD_Accept_connection - Accepts a connection and creates a new file
  descriptor

  Input Parameters:
. fd - File descriptor created by 'PX_FD_Create_network_socket' or
  'PX_FD_Create_local_socket'.

  Returns:
  New file descriptor for a new socket connection.  This file descriptor is
  marked as non-blocking.
@*/
int PX_FD_Accept_connection( fd, sock_info )
int        fd;
void       *sock_info;
{
int                newfd,
                   flags;
PX_Socket_info     *sock = (PX_Socket_info *)sock_info;
int                optval = 1;


if (sock->sock_type == AF_INET) {
    newfd = accept( fd, &sock->s1, &sock->slen );
    if (newfd >= 0) {
	setsockopt( newfd, IPPROTO_TCP, TCP_NODELAY, 
		   (char *) &optval, sizeof(optval));
	}
    }
else if (sock->sock_type == AF_UNIX)
    newfd = accept( fd, &sock->s2, &sock->slen );
else 
    newfd = -1;
if (newfd < 0) return newfd;

/* Mark this fd as non-blocking */
flags = fcntl( newfd, F_GETFL, 0 );
if (flags >= 0) {
    flags |= O_NDELAY;
    fcntl( newfd, F_SETFL, flags );
    }

return newfd;
}

/*@
  PX_FD_Create_network_conn - Creates a connection to a remote process

  Input Parameters:
. server_hostname - Network name of remote processor
. portnum - port number.
. q_wait - If true, wait for the connection to be established.

  Output Parameter:
. is_ready - Set to true if the connection is established.  May be null
  if 'q_wait' is true.

  Returns:
  A file descriptor on which a 'connect' has been `started`.  If 'q_wait' is 
  false, this 
  file descriptor is ready `only` when a 'PX_FD_Establish_network_conn'
  has completed on it, unless 'is_ready' is true.

  Notes:
  This routine uses 'gethostbyname' to get information about 'server_name';
  as a result, this routine is not thread-safe (because 'gethostbyname' is
  not thread safe), and should not be used in a signal handler.
@*/
int PX_FD_Create_network_conn( server_hostname, portnum, q_wait, is_ready )
char *server_hostname;
int  portnum, q_wait, *is_ready;
{
struct hostent     *hp;
struct sockaddr_in sa;
int                fd;
int                optval = 1;
int                flags;

hp = gethostbyname( server_hostname );
if (!hp) {
    return -1;
    }

bzero( &sa, sizeof(sa) );
bcopy( hp->h_addr, &sa.sin_addr, hp->h_length);
sa.sin_family = hp->h_addrtype;
sa.sin_port   = htons( (u_short) portnum );

fd = socket( AF_INET, SOCK_STREAM, TCP );
if (fd < 0) {
    return -1;
    }

if (setsockopt( fd, IPPROTO_TCP, TCP_NODELAY, 
	        (char *)&optval, sizeof(optval) )) {
    perror( "Error calling setsockopt:" );
    }

/* If a non-blocking socket, then can use select on write to test for
   connect would succeed.  Thus, we mark the socket as non-blocking now */
/* Mark this fd as non-blocking */
if (!q_wait) {
    flags = fcntl( fd, F_GETFL, 0 );
    if (flags >= 0) {
	flags |= O_NDELAY;
	fcntl( fd, F_SETFL, flags );
	}
    }

if (is_ready) *is_ready = 1;
if (connect( fd, &sa, sizeof(sa) ) < 0) {
    switch (errno) {
	case ECONNREFUSED:
	/* (close socket, get new socket, try again) */
	if (q_wait)
	    close(fd);
	return -1;

	case EINPROGRESS: /*  (nonblocking) - select for writing. */
	if (is_ready) *is_ready = 0;
	break;

	case EISCONN: /*  (already connected) */
	break;

	case ETIMEDOUT: /* timed out */
	return -1;

	default:
	return -1;
	}
    }
if (fd >= 0 && q_wait) {
    flags = fcntl( fd, F_GETFL, 0 );
    if (flags >= 0) {
	flags |= O_NDELAY;
	fcntl( fd, F_SETFL, flags );
	}
    }

return fd;
}

/*@
  PX_FD_Create_local_conn - Creates a connection to a local process

  Input Parameters:
. server_path - Path name to socket
. q_wait - If true, wait for the connection to be established.

  Output Parameter:
. is_ready - Set to true if the connection is established.  May be null
  if 'q_wait' is true.

  Returns:
  A file descriptor on which a 'connect' has been `started`.  If 'q_wait' is 
  false, this 
  file descriptor is ready `only` when a 'PX_FD_Establish_local_conn'
  has completed on it, unless 'is_ready' is true.
@*/
int PX_FD_Create_local_conn( server_path, q_wait, is_ready )
char *server_path;
int  q_wait, *is_ready;
{
struct sockaddr_un sa;
int                fd;
int                flags;

bzero( &sa, sizeof(sa) );
sa.sun_family = AF_UNIX;
strncpy( sa.sun_path, server_path, sizeof(sa.sun_path) - 1 );

fd = socket( AF_UNIX, SOCK_STREAM, TCP );
if (fd < 0) {
    return -1;
    }

/* If a non-blocking socket, then can use select on write to test for
   connect would succeed.  Thus, mark this fd as non-blocking */
if (!q_wait) {
    flags = fcntl( fd, F_GETFL, 0 );
    if (flags >= 0) {
	flags |= O_NDELAY;
	fcntl( fd, F_SETFL, flags );
	}
    }

if (is_ready) *is_ready = 1;
if (connect( fd, &sa, sizeof(sa) ) < 0) {
    perror( "Failed to connect: " );
    switch (errno) {
	case ECONNREFUSED:
	/* (close socket, get new socket, try again) */
	return -1;

	case EINPROGRESS: /*  (nonblocking) - select for writing. */
	if (is_ready) *is_ready = 0;
	break;
	
	case EISCONN: /*  (already connected) */
	break;
	
	case ETIMEDOUT: /* timed out */
	return -1;

	default:
	return -1;
	}
    }

if (fd >= 0 && q_wait) {
    flags = fcntl( fd, F_GETFL, 0 );
    if (flags >= 0) {
	flags |= O_NDELAY;
	fcntl( fd, F_SETFL, flags );
	}
    }

return fd;
}

/*
 * It isn't clear from the manual whether a new connect is needed or not,
 * or how you get error info the connection fails.  I'm going to
 * assume for now that the "write" signal is all you get
 */
/*@
  PX_FD_Establish_network_conn - Completes the establishment of a connection
  
  Input Parameter:
. fd - File descriptor for connection to complete.

  Notes:
  This routine should be called only if 'PX_FD_Create_network_conn' returned
  false for 'is_ready' `and` 'PX_Gselect' has determined that 'fd' is
  available for `writing`.
@*/
int PX_FD_Establish_network_conn( fd )
int fd;
{
return fd;
}

/*@
  PX_FD_Establish_local_conn - Completes the establishment of a connection
  
  Input Parameter:
. fd - File descriptor for connection to complete.

  Notes:
  This routine should be called only if 'PX_FD_Create_local_conn' returned
  false for 'is_ready' `and` 'PX_Gselect' has determined that 'fd' is
  available for `writing`.
@*/
int PX_FD_Establish_local_conn( fd )
int fd;
{
return fd;
}

/*@
  PX_FD_Free_sockinfo - Frees sockinfo information

  Input Parameter:
. sock_info - socket inforamation returned by 'PX_FD_Create_network_socket'
  or 'PX_FD_Create_local_socket'.

@*/
int PX_FD_Free_sockinfo( sock_info )
void *sock_info;
{
FREE( sock_info );
return 0;
}
